# -*- coding: utf-8 -*-


import pkg_resources
from glob import glob
from tempfile import mkdtemp
from os import sep, readlink, environ, uname
from os.path import basename, dirname, exists, getsize, isdir, isfile, islink, join, pardir

import utils
from utils import public
from ui_widgets import DialogError, DialogConfirm, DialogGetKey, SelectItem, \
                       DialogEntry, DialogDoubleEntry, DialogFindGrep, TreeView, \
                       DialogPerms, DialogOwner, InternalView
from common import *


########################################################################
##### Module variables
app = None


########################################################################
##### Action dispatcher
def do(mapp, action):
    global app
    app = mapp
    log.debug('Execute: {}'.format(action))
    try:
        fn = globals()[action]
    except KeyError:
        log.warning('ERROR can\'t execute an undefined function: {}()'.format(action))
        return RET_NONE
    if not utils.is_public_api(fn):
        log.warning('ERROR: function {}() is not public API for action "{}"'.format(fn.__name__, action))
        return RET_NONE
    ret = fn()
    log.debug('Returns: {}'.format(ret))
    return ret


########################################################################
##### Cursor movement
@public
def cursor_up():
    app.pane_active.tab_active.i -= 1
    return RetCode.fix_limits, None

@public
def cursor_down():
    app.pane_active.tab_active.i += 1
    return RetCode.fix_limits, None

@public
def cursor_up10():
    app.pane_active.tab_active.i -= 10
    return RetCode.fix_limits, None

@public
def cursor_down10():
    app.pane_active.tab_active.i += 10
    return RetCode.fix_limits, None

@public
def cursor_pageup():
    app.pane_active.tab_active.i -= app.pane_active.fh
    return RetCode.fix_limits, None

@public
def cursor_pagedown():
    app.pane_active.tab_active.i += app.pane_active.fh
    return RetCode.fix_limits, None

@public
def cursor_home():
    app.pane_active.tab_active.i = 0
    return RetCode.fix_limits, None

@public
def cursor_end():
    app.pane_active.tab_active.i = app.pane_active.tab_active.n-1
    return RetCode.fix_limits, None

@public
def cursor_goto_file():
    text = DialogEntry('Go to file', 'Type part of the file name').run()
    if not text:
        return RetCode.nothing, None
    start = app.pane_active.tab_active.i + 1
    entries = app.pane_active.tab_active.fs.get_filenames(start=start)
    for e in entries:
        if e.find(text) != -1:
            app.pane_active.tab_active.i = entries.index(e) + start
            return RetCode.fix_limits, None
    return RetCode.fix_limits, None

@public
def cursor_goto_file_1char():
    start = app.pane_active.tab_active.i + 1
    entries = app.pane_active.tab_active.fs.get_filenames(start=start)
    app.win.nodelay(0)
    ch = app.win.getkey()
    app.win.nodelay(1)
    for e in entries:
        if ch == e[0]:
            app.pane_active.tab_active.i = entries.index(e) + start
            return RetCode.fix_limits, None
    return RetCode.nothing, None


########################################################################
##### Cursor movement
@public
def dir_up():
    tab = app.pane_active.tab_active
    if tab.fs.path_str == sep:
        return RetCode.nothing, None
    olddirname = tab.fs.basename
    tab.goto_folder(tab.fs.dirname)
    tab.focus_file(olddirname)
    return RetCode.full_redisplay, None

@public
def dir_enter():
    tab = app.pane_active.tab_active
    fname = tab.fs[tab.i].name
    if fname == pardir:
        return dir_up()
    ft = tab.fs[tab.i].get_type_from_ext(app.cfg.files_ext)
    if ft in ('dir', 'archive'):
        tab.goto_folder(join(tab.fs.path_str, fname))
    elif ft in ('audio', 'ebook', 'graphics', 'pdf', 'video', 'web'):
        utils.run_on_current_file(app.cfg.programs[ft], app.pane_active.tab_active.current_filename_full, True)
        return refresh()
    else:
        utils.run_on_current_file(app.cfg.programs['pager'], app.pane_active.tab_active.current_filename_full)
        return refresh()
    return RetCode.full_redisplay, None

@public
def goto():
    newdir = DialogEntry('Change directory', 'Type directory name', '',
                         history=app.history['path'][:], is_files=True).run()
    if not newdir:
        return RetCode.nothing, None
    newdir = utils.get_norm_path(newdir, app.pane_active.tab_active.dirname)
    log.debug('Chdir "{}"'.format(newdir))
    app.pane_active.tab_active.goto_folder(newdir, delete_vfs_tree=True)
    if newdir:
        app.history.append('path', newdir)
    return RetCode.full_redisplay, None

@public
def bookmark_goto():
    while True:
        key = DialogGetKey('Goto bookmark',
                           'Press 0-9 a-z to select the bookmark, Ctrl-C to quit')
        if key == -1: # Ctrl-C
            break
        elif chr(key) in BOOKMARKS_KEYS:
            log.debug('Goto bookmark in key "{}" > "{}"'.format(chr(key), app.cfg.bookmarks[chr(key)]))
            app.pane_active.tab_active.goto_folder(app.cfg.bookmarks[chr(key)], delete_vfs_tree=True)
            break
    return RetCode.full_redisplay, None

@public
def bookmark_set():
    fs = app.pane_active.tab_active.fs
    if fs.vfs:
        DialogError('Cannot save a bookmark to a VFS')
        return RetCode.full_redisplay, None
    while True:
        key = DialogGetKey('Set bookmark',
                           'Press 0-9 a-z to save bookmark, Ctrl-C to quit')
        if key == -1: # Ctrl-C
            break
        elif chr(key) in BOOKMARKS_KEYS:
            log.debug('Save bookmark "{}" to key "{}"'.format(fs.path_str, chr(key)))
            app.cfg.bookmarks[chr(key)] = fs.path_str
            break
    return RetCode.full_redisplay, None

@public
def bookmark_select_fromlist():
    bmks = sorted(['{}  {}'.format(k, b) for k, b in app.cfg.bookmarks.items()])
    ret = SelectItem('Select Bookmark', bmks).run()
    if ret != -1:
        app.pane_active.tab_active.goto_folder(ret[3:], delete_vfs_tree=True)
    return RetCode.full_redisplay, None

@public
def history_select_fromlist():
    tab = app.pane_active.tab_active
    if len(tab.history) == 0:
        DialogError('No entries in history')
        return RetCode.full_redisplay, None
    ret = SelectItem('Return to', list(reversed(tab.history)), quick_key=False).run()
    if ret != -1:
        app.pane_active.tab_active.goto_folder(ret, delete_vfs_tree=True)
    return RetCode.full_redisplay, None


########################################################################
##### Panes
@public
def pane_change_focus():
    otherpane = app.pane2 if app.pane1.focus else app.pane1
    app.focus_pane(otherpane)
    return RetCode.full_redisplay, None

@public
def pane_other_tab_equal():
    path = app.pane_active.tab_active.fs.path_str
    app.pane_inactive.tab_active.goto_folder(path, delete_vfs_tree=True)
    return RetCode.full_redisplay, None

@public
def panes_swap():
    log.debug('Swap panes')
    app.pane1, app.pane2 = app.pane2, app.pane1
    app.resize()
    return RetCode.full_redisplay, None

@public
def panes_cycle_view():
    if app.pane_active.mode == PaneMode.half:
        app.pane_active.change_mode(PaneMode.full)
    else:
        app.pane_active.change_mode(PaneMode.half)
    return RetCode.full_redisplay, None

@public
def refresh():
    for tab in app.pane1.tabs + app.pane2.tabs:
        tab.reload()
        tab.refresh()
    return RetCode.full_redisplay, None

@public
def redraw_screen():
    app.clear_screen()
    return RetCode.full_redisplay, None

@public
def dotfiles_toggle():
    app.cfg.options.show_dotfiles = not app.cfg.options.show_dotfiles
    for tab in app.pane1.tabs + app.pane2.tabs:
        tab.fs.cfg.show_dotfiles = app.cfg.options.show_dotfiles
        tab.refresh()
    return RetCode.full_redisplay, None

@public
def filters_edit():
    tab = app.pane_active.tab_active
    text = DialogEntry('Edit filter', 'Type globs, separated by commas, for the files you want to hide',
                       tab.fs.cfg.filters, history=app.history['glob'][:], is_files=False).run()
    if text is None: # not "if not text"!!! because we need text='' to clear filters
        return RetCode.nothing, None
    tab.fs.cfg.filters = text
    tab.refresh()
    if text:
        app.history.append('glob', text)
    return RetCode.half_redisplay, None

@public
def sort_files():
    sorttypes = {'o': SortType.none, 'O': SortType.none,
                 'n': SortType.byName, 'N': SortType.byName,
                 'e': SortType.byName, 'E': SortType.byName,
                 's': SortType.bySize, 'S': SortType.bySize,
                 'd': SortType.byMTime, 'D': SortType.byMTime}
    while True:
        key = DialogGetKey('Sorting mode',
                           'N(o)ne, by (n)ame, by (e)xtension, by (s)ize, by (d)ate,\nuppercase to reverse order, Ctrl-C to quit')
        if key == -1: # Ctrl-C
            break
        elif chr(key) in sorttypes.keys():
            app.cfg.options.sort_type = sorttypes[chr(key)]
            app.cfg.options.sort_reverse = key<ord('Z')
            app.pane1.refresh()
            app.pane2.refresh()
            break
    return RetCode.full_redisplay, None

@public
def show_dirs_size():
    tab = app.pane_active.tab_active
    dirs = [e for e in tab.selected if e.is_dir] if tab.selected else tab.fs.dirs
    tab.selected = []
    if len(dirs) == 0:
        return RetCode.nothing, None
    log.debug('Show directories size')
    args = [(d.pfile, ) for d in dirs]
    st, rets, errs = utils.ProcessFuncLoop('Calculate directories size', utils.get_dir_size, args).run()
    if st == ProcCode.stopped:
        return RetCode.full_redisplay, None
    for d, size in zip(dirs, rets):
        if size != -1:
            tab.fs.lookup(basename(d.pfile)).update_size(size)
    return RetCode.full_redisplay, None


@public
def tab_new():
    if len(app.pane_active.tabs) >= MAX_TABS:
        DialogError('Cannot create more tabs')
    else:
        curtab = app.pane_active.tab_active
        app.pane_active.insert_new_tab(curtab.fs.path_str, curtab)
    return RetCode.full_redisplay, None

@public
def tab_close():
    if len(app.pane_active.tabs) == 1:
        DialogError('Cannot close last tab')
    else:
        app.pane_active.close_tab(app.pane_active.tab_active)
    return RetCode.full_redisplay, None

@public
def tab_left():
    tab = app.pane_active.tab_active
    idx = app.pane_active.tabs.index(tab)
    if idx == 0:
        return RetCode.nothing, None
    app.pane_active.tab_active = app.pane_active.tabs[idx-1]
    return RetCode.full_redisplay, None

@public
def tab_right():
    tab = app.pane_active.tab_active
    idx = app.pane_active.tabs.index(tab)
    if idx==len(app.pane_active.tabs)-1 or idx==MAX_TABS-1:
        return RetCode.nothing, None
    app.pane_active.tab_active = app.pane_active.tabs[idx+1]
    return RetCode.full_redisplay, None


########################################################################
##### Selections
@public
def select():
    tab = app.pane_active.tab_active
    if tab.i != 0: # pardir
        it = tab.fs[tab.i]
        if it is not None:
            try:
                tab.selected.index(it)
            except ValueError:
                tab.selected.append(it)
            else:
                tab.selected.remove(it)
            app.pane_active.tab_active.i += 1
            return RetCode.fix_limits, None
    return RetCode.nothing, None

@public
def select_glob():
    tab = app.pane_active.tab_active
    text = DialogEntry('Select group', 'Type pattern', '*', history=app.history['glob'][:], is_files=False).run()
    if not text:
        return RetCode.nothing, None
    for fname in glob(join(tab.dirname, text)):
        f = tab.fs.lookup(basename(fname))
        if f and f not in tab.selected:
            tab.selected.append(f)
    if text != '*':
        app.history.append('glob', text)
    return RetCode.half_redisplay, None

@public
def deselect_glob():
    tab = app.pane_active.tab_active
    text = DialogEntry('Deselect group', 'Type pattern', '*', history=app.history['glob'][:], is_files=False).run()
    if not text:
        return RetCode.nothing, None
    for fname in glob(join(tab.dirname, text)):
        f = tab.fs.lookup(basename(fname))
        if f and f in tab.selected:
            tab.selected.remove(f)
    if text != '*':
        app.history.append('glob', text)
    return RetCode.half_redisplay, None

@public
def select_invert():
    tab = app.pane_active.tab_active
    selected_old = tab.selected[:]
    tab.selected = [f for f in tab.fs if f not in selected_old and f.name != '..']
    return RetCode.half_redisplay, None


########################################################################
##### Files
@public
def rename_file():
    tab = app.pane_active.tab_active
    filename = tab.current_filename
    if filename == pardir:
        return RetCode.nothing, None
    newname = DialogEntry('Rename', 'Rename \'{}\' to'.format(filename), filename,
                          history=app.history['file'][:], is_files=True).run()
    if not newname:
        return RetCode.nothing, None
    src = tab.current_filename_full
    dest = newname if newname[0] == sep else join(tab.dirname, newname)
    if src == dest:
        DialogError('Cannot rename \'{}\'\nSource and destination are the same file'.format(filename))
        return RetCode.full_redisplay, None
    if dirname(dest) != tab.dirname:
        DialogError('Cannot rename to different directory')
        return RetCode.full_redisplay, None
    if exists(dest) and app.cfg.confirmations.overwrite:
        if not DialogConfirm('Rename file', 'Overwrite \'{}\'?'.format(newname), 0):
            return RetCode.full_redisplay, None
    log.debug('Rename file: {} -> {}'.format(src, dest))
    err = utils.rename_file(src, dest)
    if err:
        log.warning('Cannot rename: \'{}\' to \'{}\': {}'.format(filename, newname, str(err)))
        DialogError('Cannot rename: \'{}\' to \'{}\'\n{}'.format(filename, newname, str(err)))
        return RetCode.full_redisplay, None
    else:
        app.history.append('file', newname)
        return refresh()

def __copymove_helper(act):
    tab = app.pane_active.tab_active
    files = [f.pfile for f in tab.selected_or_current]
    if len(files) == 0:
        return None, None, None
    if len(files) == 1:
        subtitle = '{} \'{}\' to:'.format(act, basename(files[0]))
    else:
        subtitle = '{} {} items to:'.format(act, len(files))
    destdir = DialogEntry('{} file(s)'.format(act), subtitle, app.pane_inactive.tab_active.dirname+sep,
                          history=app.history['path'][:], is_files=True).run()
    if not destdir:
        return None, None, None
    destdir = utils.get_norm_path(destdir, tab.dirname)
    if not isdir(destdir) and len(files) > 1:
        DialogError('Cannot {0} files\nTried to {0} many items to one file name'.format(act.lower()))
        return None, None, None
    log.debug('{} file(s) to \'{}\''.format(act, destdir))
    app.history.append('path', destdir)
    basepath = tab.dirname if tab.dirname[-1]==sep else tab.dirname+sep
    return destdir, basepath, files

@public
def copy_file():
    destdir, basepath, files = __copymove_helper('Copy')
    if not destdir:
        return RetCode.full_redisplay, None
    es = utils.PathContents(files, basepath)
    args = [(f, s, e, basepath, destdir) for f, s, e in es.entries]
    st, rets, errs = utils.ProcessFuncCopyLoop('Copy file(s)', utils.copy_file, args,
                                               es.tsize, app.cfg.confirmations.overwrite).run()
    app.pane_active.tab_active.selected = []
    if st == ProcCode.stopped:
        return refresh()
    for a, res, err in zip(args, rets, errs):
        if err:
            if isinstance(err, LFMFileSkipped):
                log.warning('Copy file(s). Not overwritten: {}'.format(str(err)))
            else:
                log.warning('ERROR: {}'.format(str(err).replace('\n', ' ')))
    return refresh()

@public
def move_file():
    destdir, basepath, files = __copymove_helper('Move')
    if not destdir:
        return RetCode.full_redisplay, None
    es = utils.PathContents(files, basepath)
    args = [(f, s, e, basepath, destdir) for f, s, e in es.entries]
    st, rets, errs = utils.ProcessFuncCopyLoop('Move file(s)', utils.copy_file, args,
                                               es.tsize, app.cfg.confirmations.overwrite).run()
    if st == ProcCode.stopped:
        return refresh()
    not_overwritten = list()
    for a, res, err in zip(args, rets, errs):
        if err:
            if isinstance(err, LFMFileSkipped):
                log.warning('Move file(s). Not overwritten: {}'.format(str(err)))
                not_overwritten.append(str(err))
            else:
                log.warning('ERROR: {}'.format(str(err).replace('\n', ' ')))
    es.remove_files(not_overwritten)
    args = [(f, s, e, basepath) for f, s, e in es.entries_rev] # reverse!
    st, rets, errs = utils.ProcessFuncDeleteLoop('Move file(s)', utils.delete_file, args,
                                                 es.tsize, False).run()
    app.pane_active.tab_active.selected = []
    if st == ProcCode.stopped:
        return refresh()
    for a, res, err in zip(args, rets, errs):
        if err:
            log.warning('ERROR: {}'.format(str(err).replace('\n', ' ')))
    return refresh()

@public
def move_file2():
    # alternative version using shtutil.move instead of copy & delete. Faster but less control
    destdir, basepath, files = __copymove_helper('Move')
    if not destdir:
        return RetCode.full_redisplay, None
    args, size = list(), 0
    for f in files:
        try:
            s = getsize(f)
        except OSError as err:
            s = 0
        size += s
        args.append((f, s, None, basepath, destdir))
    st, rets, errs = utils.ProcessFuncCopyLoop('Move file(s)', utils.move_file, args, size,
                                               app.cfg.confirmations.overwrite).run()
    app.pane_active.tab_active.selected = []
    if st == ProcCode.stopped:
        return refresh()
    for a, ret, err in zip(args, rets, errs):
        if err is not None:
            log.warning('ERROR: {}'.format(str(err).replace('\n', ' ')))
    return refresh()

@public
def delete_file():
    tab = app.pane_active.tab_active
    files = [f.pfile for f in tab.selected_or_current]
    if len(files) == 0:
        return RetCode.nothing, None
    log.debug('Delete file(s)')
    es = utils.PathContents(files, tab.dirname+sep)
    args = [(f, s, e, tab.dirname+sep) for f, s, e in es.entries_rev] # reverse!
    st, rets, errs = utils.ProcessFuncDeleteLoop('Delete file(s)', utils.delete_file, args,
                                                 es.tsize, app.cfg.confirmations.delete).run()
    app.pane_active.tab_active.selected = []
    if st == ProcCode.stopped:
        return refresh()
    for a, res, err in zip(args, rets, errs):
        if err:
            if isinstance(err, LFMFileSkipped):
                log.warning('Delete file(s). Not deleted: {}'.format(str(err)))
            else:
                log.warning('ERROR: {}'.format(str(err).replace('\n', ' ')))
    return refresh()

@public
def exec_on_file():
    cmd = DialogEntry('Execute command on file(s)', 'Enter command', '',
                      history=app.history['exec'][:], is_files=False).run()
    if not cmd:
        return RetCode.nothing, None
    for fsentry in app.pane_active.tab_active.selected_or_current:
        log.debug('Exec on file: {} "{}"'.format(cmd, fsentry.pfile))
        utils.run_on_current_file(cmd, fsentry.pfile)
    log.debug('Exec on file: {}'.format(cmd))
    app.pane_active.tab_active.selected = []
    app.history.append('exec', cmd)
    return refresh()

@public
def view_file():
    log.debug('View file: {}'.format(app.pane_active.tab_active.current_filename_full))
    utils.run_on_current_file(app.cfg.programs['pager'], app.pane_active.tab_active.current_filename_full)
    return refresh()

@public
def edit_file():
    log.debug('Edit file: {}'.format(app.pane_active.tab_active.current_filename_full))
    utils.run_on_current_file(app.cfg.programs['editor'], app.pane_active.tab_active.current_filename_full)
    return refresh()

@public
def make_dir():
    newdir = DialogEntry('Make directory', 'Type directory name', '',
                         history=app.history['file'][:], is_files=True).run()
    if not newdir:
        return RetCode.nothing, None
    log.debug('Make directory: {}'.format(join(app.pane_active.tab_active.dirname, newdir)))
    err = utils.make_dir(join(app.pane_active.tab_active.dirname, newdir))
    if err:
        log.warning('Cannot make directory: {}'.format(str(err)))
        DialogError('Cannot make directory\n{}'.format(str(err)))
        return RetCode.full_redisplay, None
    else:
        app.history.append('file', newdir)
        return refresh()

@public
def touch_file():
    filename = DialogEntry('Touch file', 'Type file name', '',
                           history=app.history['file'][:], is_files=True).run()
    if not filename:
        return RetCode.nothing, None
    log.debug('Touch file: {}'.format(join(app.pane_active.tab_active.dirname, filename)))
    err = utils.touch_file(join(app.pane_active.tab_active.dirname, filename))
    if err:
        log.warning('Cannot touch file: {}'.format(str(err)))
        DialogError('Cannot touch file\n{}'.format(str(err)))
        return RetCode.full_redisplay, None
    else:
        app.history.append('file', filename)
        return refresh()

@public
def link_create():
    thistab, othertab = app.pane_active.tab_active, app.pane_inactive.tab_active
    otherfile = utils.get_relpath(join(othertab.dirname, othertab.current_filename), thistab.dirname)
    ans = DialogDoubleEntry('Create link', 'Link name', 'Pointing to', '', otherfile,
                            history1=app.history['file'][:], history2=app.history['path'][:],
                            is_files=True).run()
    if not ans:
        return RetCode.nothing, None
    newlink, pointto = ans
    if newlink == '':
        DialogError('Cannot create link\nYou must specify the name for the new link')
        return RetCode.full_redisplay, None
    if pointto == '':
        DialogError('Cannot create link\nYou must specify the file to link')
        return RetCode.full_redisplay, None
    log.debug('Create link: {} -> {}'.format(join(thistab.dirname, newlink), pointto))
    err = utils.link_create(join(thistab.dirname, newlink), pointto)
    if err:
        log.warning('Cannot create link: {}'.format(str(err)))
        DialogError('Cannot create link\n{}'.format(str(err)))
        return RetCode.full_redisplay, None
    else:
        app.history.append('file', newlink)
        return refresh()

@public
def link_edit():
    linkname = app.pane_active.tab_active.current_filename_full
    if not islink(linkname):
        return RetCode.nothing, None
    newpointto = DialogEntry('Edit link', 'Link points to', readlink(linkname),
                             history=app.history['path'][:], is_files=True).run()
    if not newpointto:
        return RetCode.nothing, None
    if newpointto == readlink(linkname):
        return RetCode.nothing, None
    log.debug('Edit link: {} -> {}'.format(linkname, newpointto))
    err = utils.link_edit(linkname, newpointto)
    if err:
        log.warning('Cannot edit link: {}'.format(str(err)))
        DialogError('Cannot edit link\n{}'.format(str(err)))
        return RetCode.full_redisplay, None
    else:
        app.history.append('path', newpointto)
        return refresh()

@public
def backup_file():
    ext = app.cfg.misc.backup_extension
    args = [(f.pfile, ext) for f in app.pane_active.tab_active.selected_or_current]
    if len(args) == 0:
        return RetCode.nothing, None
    log.debug('Backup file(s)')
    st, rets, errs = utils.ProcessFuncLoop('Create backup file(s)', utils.backup_file, args).run()
    app.pane_active.tab_active.selected = []
    if st == ProcCode.stopped:
        return refresh()
    for a, ret, err in zip(args, rets, errs):
        if err is not None:
            log.warning('ERROR: {}'.format(str(err).replace('\n', ' ')))
    return refresh()

@public
def change_perms():
    tab = app.pane_active.tab_active
    files = [f for f in tab.selected_or_current]
    if len(files) == 0:
        return RetCode.nothing, None
    log.debug('Change file permissions')
    app.pane_active.tab_active.selected = []
    for_all = False
    for i, f in enumerate(files):
        if not for_all:
            perms, recursive, for_all = DialogPerms(f.name, f.mode, i+1, len(files)).run()
            if perms == -1: # stop
                break
            elif perms == 0: # ignore file
                continue
        cmd = 'chmod {} 0{:o} "{}"'.format('-R' if recursive else '', utils.str2perms(perms), f.pfile)
        st, res, err = utils.ProcessCommand('Change file(s) permission', cmd, cmd, path=tab.dirname).run()
        if st == -100: # stopped by user
            return RetCode.full_redisplay, None
        if err:
            log.warning('Cannot "{}": {}'.format(cmd, str(err).replace('\n', ' ')))
            DialogError('Cannot "{}"\n{}'.format(cmd, str(err)))
    return refresh()

@public
def change_owner():
    tab = app.pane_active.tab_active
    files = [f for f in tab.selected_or_current]
    if len(files) == 0:
        return RetCode.nothing, None
    log.debug('Change file owner/group')
    app.pane_active.tab_active.selected = []
    for_all = False
    for i, f in enumerate(files):
        if not for_all:
            owner, group, recursive, for_all = DialogOwner(f.name, f.owner_str, f.group_str,
                                                           utils.get_owners(), utils.get_groups(),
                                                           i+1, len(files)).run()
            if owner == -1: # stop
                break
            elif owner == 0: # ignore file
                continue
        rec = '-R' if recursive else ''
        cmd = 'chown {} {} "{}"'.format(rec, owner, f.pfile)
        st, res, err = utils.ProcessCommand('Change file(s) owner', cmd, cmd, path=tab.dirname).run()
        if st == -100: # stopped by user
            return RetCode.full_redisplay, None
        if err:
            log.warning('Cannot "{}": {}'.format(cmd, str(err).replace('\n', ' ')))
            DialogError('Cannot "{}"\n{}'.format(cmd, str(err)))
        cmd = 'chgrp {} {} "{}"'.format(rec, group, f.pfile)
        st, res, err = utils.ProcessCommand('Change file(s) group', cmd, cmd, path=tab.dirname).run()
        if st == -100: # stopped by user
            return RetCode.full_redisplay, None
        if err:
            log.warning('Cannot "{}": {}'.format(cmd, str(err).replace('\n', ' ')))
            DialogError('Cannot "{}"\n{}'.format(cmd, str(err)))
    return refresh()

@public
def diff_file_with_backup():
    log.debug('Diff file with backup')
    ext = app.cfg.misc.backup_extension
    filename = app.pane_active.tab_active.current_filename_full
    if filename.endswith(ext):
        file_old, file_new = filename, filename[:-len(ext)]
    else:
        file_old, file_new = filename+ext, filename
    if not exists(file_old):
        DialogError('Cannot diff file\nBackup file does not exist')
        return RetCode.full_redisplay, None
    if not exists(file_new):
        DialogError('Cannot diff file\nOnly backup file exists')
        return RetCode.full_redisplay, None
    if not isfile(file_old) or not isfile(file_new):
        DialogError('Cannot diff file\nWe can only diff regular files')
        return RetCode.full_redisplay, None
    diff = utils.get_file_diff(file_old, file_new, app.cfg.misc.diff_type)
    if diff == '':
        DialogError('Files are identical')
        return RetCode.full_redisplay, None
    tmpfile = join(mkdtemp(), basename(file_old) + ' DIFF ' + basename(file_new))
    with open(tmpfile, 'w') as f:
        f.write(diff)
    utils.run_on_current_file(app.cfg.programs['pager'], tmpfile)
    utils.delete_bulk(tmpfile, True)
    return RetCode.full_redisplay, None

@public
def show_file_info():
    log.debug('Show file information')
    tab = app.pane_active.tab_active
    f = tab.current
    lst = []
    user = environ['USER']
    username = utils.get_user_fullname(user)
    so, host, ver, tmp, arch = uname()
    color = 'view_green_on_black'
    lst.append(('{} v{} executed by {}'.format(LFM_NAME, VERSION, username), color))
    lst.append(('<{}@{}> on {} {} [{}]'.format(user, host, so, ver, arch), color))
    lst.append(('', color))
    color = 'view_red_on_black'
    fileinfo = utils.get_file_info(f.pfile)
    lst.append(('{}: {} ({})'.format(FILETYPES[f.type][1], f.name, fileinfo), color))
    lst.append(('Path: {}'.format(tab.fs.path_str), color))
    lst.append(('Size: {} bytes'.format(f.size), color))
    lst.append(('Mode: {} ({:o})'.format(f.mode_str, f.mode), color))
    lst.append(('Links: {}'.format(f.stat.st_nlink), color))
    lst.append(('User ID: {} ({}) / Group ID: {} ({})'.format(f.owner_str, f.owner, f.group_str, f.group), color))
    lst.append(('Last access: {}'.format(utils.time2str_full(f.stat.st_atime)), color))
    lst.append(('Last modification: {}'.format(utils.time2str_full(f.mtime)), color))
    lst.append(('Last change: {}'.format(utils.time2str_full(f.stat.st_ctime)), color))
    dev = f.stat.st_dev
    lst.append(('Location: {}, {} / Inode: #{:X} ({:X}h:{:X}h)'.format(
        (dev>>8) & 0x00FF, dev & 0x00FF, f.stat.st_ino, dev, f.stat.st_ino), color))
    filename = tab.fs.path_str if tab.fs.vfs else f.pfile
    mountpoint, device, fstype = utils.get_mountpoint_for_file(filename)
    lst.append(('File system: {} on {} ({})'.format(device, mountpoint, fstype), color))
    InternalView('Information about \'{}\''.format(f.name), lst).run()
    return refresh()

@public
def show_filesystems_info():
    log.debug('Show filesystems information')
    try:
        buf = utils.get_filesystems_info()
    except OSError as err:
        DialogError('Cannot show filesystems info:\n{}'.format(err))
        return RetCode.full_redisplay, None
    lbuf = buf.strip().split('\n')
    hdr = lbuf[0].strip()
    lst = [(hdr, 'view_red_on_black'), ('-'*len(hdr), 'view_red_on_black')]
    for l in lbuf[1:]:
        lst.append((l.strip(), 'view_white_on_black'))
    InternalView('Show filesystems info', lst).run()
    return refresh()


########################################################################
##### Un/compress
def do_uncompress_dir(destdir):
    args = [(f.pfile, destdir) for f in app.pane_active.tab_active.selected_or_current]
    if len(args) == 0:
        return RetCode.nothing, None
    log.debug('Uncompress file(s) to \'{}\''.format(destdir))
    app.pane_active.tab_active.selected = []
    st, rets, errs = utils.ProcessFuncLoop('Uncompress file(s)', utils.uncompress_dir, args).run()
    if st == ProcCode.stopped:
        return refresh()
    for a, res, err in zip(args, rets, errs):
        if err:
            log.warning('ERROR: {}'.format(str(err).replace('\n', ' ')))
    return refresh()

@public
def uncompress_dir():
    return do_uncompress_dir(app.pane_active.tab_active.dirname)

@public
def uncompress_dir_other_pane():
    return do_uncompress_dir(app.pane_inactive.tab_active.dirname)

@public
def compress_dir():
    dirs = [d.pfile for d in app.pane_active.tab_active.selected_or_current] # if d.is_dir
    if len(dirs) == 0:
        return RetCode.nothing, None
    compress_fmts = {'g': 'tgz', 'b': 'tbz2', 'x': 'txz', 'l': 'tlz', '4': 'tlz4',
                     't': 'tar', 'z': 'zip', 'r': 'rar', '7': '7z'}
    while True:
        ch = DialogGetKey('Compress directory to...',
                          '.tar.(g)z, .tar.(b)z2, .tar.(x)z, .tar.(l)z,\n.tar.lz(4), .(t)ar, .(z)ip, .(r)ar, .(7)z\n\n'
                          '                Ctrl-C to quit')
        if ch == -1: # Ctrl-C
            return RetCode.full_redisplay, None
        elif chr(ch) in compress_fmts.keys():
            typ = compress_fmts[chr(ch)]
            break
    args = [(d, typ) for d in dirs]
    log.debug('Compress dir(s)')
    app.pane_active.tab_active.selected = []
    st, rets, errs = utils.ProcessFuncLoop('Compress dir(s)', utils.compress_dir, args).run()
    if st == ProcCode.stopped:
        return refresh()
    for a, res, err in zip(args, rets, errs):
        if err:
            log.warning('ERROR: {}'.format(str(err).replace('\n', ' ')))
    return refresh()

@public
def compress_uncompress_file():
    files = [f.pfile for f in app.pane_active.tab_active.selected_or_current] # if f.is_dir
    if len(files) == 0:
        return RetCode.nothing, None
    compress_fmts = {'g': 'gz', 'b': 'bz2', 'x': 'xz', 'l': 'lz', '4': 'lz4'}
    while True:
        ch = DialogGetKey('Un/Compress file(s)',
                          '(g)zip, (b)zip2, (x)z, (l)z, lz(4)\n\n'
                          '   Ctrl-C to quit')
        if ch == -1: # Ctrl-C
            return RetCode.full_redisplay, None
        elif chr(ch) in compress_fmts.keys():
            typ = compress_fmts[chr(ch)]
            break
    args = [(f, typ) for f in files]
    log.debug('Compress or Uncompress file(s)')
    app.pane_active.tab_active.selected = []
    st, rets, errs = utils.ProcessFuncLoop('Compress dir(s)', utils.compress_uncompress_file, args).run()
    if st == ProcCode.stopped:
        return refresh()
    for a, res, err in zip(args, rets, errs):
        if err:
            log.warning('ERROR: {}'.format(str(err).replace('\n', ' ')))
    return refresh()


########################################################################
##### General
@public
def find_grep():
    ans = DialogDoubleEntry('Find files', 'Filename', 'Content', '*', '',
                            history1=app.history['find'][:], history2=app.history['grep'][:],
                            is_files=False).run()
    if ans is None or not ans[0]:
        return RetCode.nothing, None
    fs, pat = ans
    tab = app.pane_active.tab_active
    path = tab.dirname if tab.dirname[-1]==sep else tab.dirname+sep
    if pat:
        log.debug('Find "{}" files with "{}" in {}'.format(fs, pat, path))
        # 1. find . -type f -iname "*.py" -exec grep -EHni PATTERN {} \;
        #    the slowest, 10x
        # 2. find . -type f -iname "*py" -print0 | xargs --null -0 grep -EHni PATTERN
        #    maybe the best choice
        # 3. grep -EHni PATTERN `find . -type f -iname "*.py"`
        #    don't like the `
        # 4. grep -REHni PATTERN --include "*.py" .
        #    the fastest, but maybe they wouldn't work on some old UNIX because of -R, --include
        cmd = '{} -R{}Hn{} "{}" --include "{}"'.format(SYSPROGS['grep'],
                                                       'E' if app.cfg.options.grep_regex else '',
                                                       'i' if app.cfg.options.grep_ignorecase else '',
                                                       pat, fs)
        st, res, err = utils.ProcessCommand('Find files', 'Searching for "{}" in files with "{}" in their name'.format(pat, fs),
                                            cmd, path=path).run()
        if st == -100: # stopped by user
            return RetCode.full_redisplay, None
        if err:
            log.warning('Cannot grep "{}" in "{}" files: {}'.format(pat, fs, str(err).replace('\n', ' ')))
            DialogError('Cannot grep "{}" in "{}" files\n{}'.format(pat, fs, str(err)))
            return RetCode.full_redisplay, None
        app.history.append('find', fs)
        app.history.append('grep', pat)
        if not res.strip():
            DialogError('Did not find "{}" in any file with "{}" in name'.format(pat, fs))
            return RetCode.full_redisplay, None
        entries = sorted([f.strip() for f in res.split('\n') if f.strip()])
        title, entry = 'Find "{}" pattern in "{}" files'.format(pat, fs), ''
    else:
        log.debug('Find "{}" files in {}'.format(fs, path))
        cmd = '{} "{}" -{}name "{}" -print'.format(SYSPROGS['find'], path,
                                                   'i' if app.cfg.options.find_ignorecase else '', fs)
        st, res, err = utils.ProcessCommand('Find files', 'Searching for files with "{}" in their name'.format(fs),
                                            cmd, path=path).run()
        if st == -100: # stopped by user
            return RetCode.full_redisplay, None
        if err:
            log.warning('Cannot find "{}" files: {}'.format(fs, str(err).replace('\n', ' ')))
            DialogError('Cannot find "{}" files\n{}'.format(fs, str(err)))
            return RetCode.full_redisplay, None
        app.history.append('find', fs)
        if not res.strip():
            DialogError('Did not find any file with "{}" in name'.format(fs))
            return RetCode.full_redisplay, None
        entries = sorted([f.strip().replace(path, '') for f in res.split('\n') if f.strip() and f!=path])
        title, entry = 'Find "{}" files'.format(fs), ''
    # common
    while True:
        ans, entry = DialogFindGrep(title, entries, entry).run()
        if pat:
            try:
                filename, lineno, _ = entry.split(':', 2)
            except ValueError: # entry = "-1 - Binary file .hg/store/data/lfm/utils.py.i matches"
                filename = entry
                if filename.startswith('Binary file'):
                    filename = filename[11:]
                if filename.endswith('matches'):
                    filename = filename[:-7]
                filename, lineno = filename.strip(), 0
        else:
            filename = entry
        pfilename = join(path, filename)
        if ans == 0:             # goto file
            tab.goto_folder(join(tab.fs.path_str, dirname(filename)))
            tab.i = tab.fs.pos(basename(filename))
            tab.fix_limits()
            return RetCode.full_redisplay, None
        elif ans == 1:           # panelize
            files = list()
            for f in entries:
                if f.startswith('Binary file'):
                    f = f[11:]
                if f.endswith('matches'):
                    f = f[:-7]
                f = f.split(':', 1)[0].strip() # FIXME: won't work if filename contains ':' chars
                if f not in files:
                    files.append(f)
            tab.goto_folder(tab.fs.path_str, files=files)
            return RetCode.full_redisplay, None
        elif ans == 2:           # view
            cmd = app.cfg.programs['pager']
            cmd = '{} +{}'.format(cmd, lineno) if pat else cmd
            utils.run_on_current_file(cmd, pfilename)
        elif ans == 3:           # edit
            cmd = app.cfg.programs['editor']
            cmd = '{} +{}'.format(cmd, lineno) if pat else cmd
            utils.run_on_current_file(cmd, pfilename)
        elif ans == 4:           # do something on file
            cmd = DialogEntry('Execute command on file(s)', 'Enter command', '',
                              history=app.history['exec'][:], is_files=False).run()
            if not cmd:
                continue
            app.history.append('exec', cmd)
            utils.run_on_current_file(cmd, pfilename)
        else:
            return refresh()

@public
def show_tree():
    tab = app.pane_active.tab_active
    log.debug('Tree in "{}"'.format(tab.dirname))
    res = TreeView(tab.dirname).run()
    if res != -1:
        tab.goto_folder(res)
    return RetCode.full_redisplay, None

@public
def main_menu():
    menu = ['/    Find/grep file(s)',
            '#    Show directories size',
            's    Sort files',
            't    Tree',
            'f    Show filesystems info',
            'o    Open shell',
            'c    Edit configuration',
            'k    Edit keys',
            'e    Edit theme',
            'h    Delete history']
    ret = SelectItem('Main Menu', menu, min_height=True).run()
    if ret != -1:
        ch = ret[0]
        if ch == '/':
            return find_grep()
        elif ch == '#':
            return show_dirs_size()
        elif ch == 's':
            return sort_files()
        elif ch == 't':
            return show_tree()
        elif ch == 'f':
            return show_filesystems_info()
        elif ch == 'o':
            return open_shell()
        elif ch == 'c':
            utils.run_on_current_file(app.cfg.programs['editor'], CONFIG_FILE)
            from ui import init_config
            app.cfg = init_config(False)
            return refresh()
        elif ch == 'k':
            utils.run_on_current_file(app.cfg.programs['editor'], KEYS_FILE)
            app.init_keys()
            return refresh()
        elif ch == 'e':
            utils.run_on_current_file(app.cfg.programs['editor'], THEME_FILE)
            app.init_colors()
            return refresh()
        elif ch == 'h':
            try:
                app.history.delete()
            except Exception as e:
                DialogError('Cannot save history file\n{}'.format(str(e)))
            return RetCode.nothing, None
        return RetCode.nothing, None
    else:
        return RetCode.nothing, None

@public
def file_menu():
    menu = ['@    Exec on file',
            'i    File info',
            'p    Change file permissions',
            'o    Change file owner and/or group',
            'a    Backup file(s)',
            'd    Diff file with backup',
            # 'l    Folder comparation',
            # 'y    Folder synchronization',
            'z    Compress/uncompress file(s)…',
            'x    Uncompress file',
            'u    Uncompress file in other panel',
            'c    Compress directory to format…']
    ret = SelectItem('File Menu', menu, min_height=True).run()
    if ret != -1:
        ch = ret[0]
        if ch == '@':
            return exec_on_file()
        elif ch == 'i':
            return show_file_info()
        elif ch == 'p':
            return change_perms()
        elif ch == 'o':
            return change_owner()
        elif ch == 'a':
            return backup_file()
        elif ch == 'd':
            return diff_file_with_backup()
        elif ch == 'l':
            pass
        elif ch == 'y':
            pass
        elif ch == 'z':
            return compress_uncompress_file()
        elif ch == 'x':
            return uncompress_dir()
        elif ch == 'u':
            return uncompress_dir_other_pane()
        elif ch == 'c':
            return compress_dir()
        return RetCode.nothing, None
    else:
        return RetCode.nothing, None

@public
def help_menu():
    menu = ['r    Readme',
            'n    News',
            't    Todo',
            'l    License',
            'k    Key bindings']
    ret = SelectItem('Help Menu', menu, min_height=True).run()
    if ret != -1:
        if ret[0] == 'k':
            from preferences import dump_keys_to_file
            dump_keys_to_file(app.keys)
            filename = DUMP_KEYS_FILE
            utils.run_on_current_file(app.cfg.programs['pager'], filename)
        else:
            docfile = {'r': 'README', 'n': 'NEWS', 't': 'TODO', 'l': 'COPYING'}.get(ret[0])
            try:
                filename = '/usr/local/share/doc/lfm/' + docfile
                with open(filename) as f:
                    buf = f.readlines()
                f.close()
                InternalView(docfile, [(l, 'view_white_on_black') for l in buf], center=False).run()
            except Exception as e:
                DialogError('Cannot read the doc!\n{}'.format(str(e)))
        return refresh()
    else:
        return RetCode.nothing, None

@public
def toggle_powercli():
    app.cli.toggle()
    app.display_statusbar_or_powercli()
    return refresh()

@public
def open_shell():
    utils.run_shell(app.cfg.programs['shell'], app.pane_active.tab_active.dirname)
    return refresh()

@public
def quit_chdir():
    return do_quit(chdir=True)

@public
def quit_nochdir():
    return do_quit(chdir=False)

def do_quit(chdir=True):
    if app.cfg.confirmations.quit:
        ch = DialogConfirm('Last File Manager', 'Quit Last File Manager?', 1)
        if ch == 0:
            return RetCode.full_redisplay, None
    fs = app.pane_active.tab_active.fs
    path = dirname(fs.base) if fs.vfs else fs.pdir
    for tab in app.pane1.tabs + app.pane2.tabs:
        tab.close()
    return RetCode.quit_chdir if chdir else RetCode.quit_nochdir, path


########################################################################



# -*- coding: utf-8 -*-


from os.path import exists
from shutil import copyfile
from configparser import ConfigParser
from collections import OrderedDict
import pickle

from utils import get_lfm_data_file_contents, ConfigParserWithComments, get_public_actions
from key_defs import key_str2bin, key_bin2str
from common import *


########################################################################
##### Default options and preferences
CONFIGFILE_HEADER = '{0} {1} Configuration File v3.x {0}'.format('#'*10, LFM_NAME)

DEF_OPTIONS = {'save_configuration_at_exit': True,
               'save_history_at_exit': True,
               'show_output_after_exec': True,
               'use_wide_chars': False,
               'rebuild_vfs': False,
               'detach_terminal_at_exec': True,
               'show_dotfiles': True,
               'sort_type': SortType.byName,
               'sort_reverse': False,
               'sort_mix_dirs': False,
               'sort_mix_cases': True,
               'automatic_file_encoding_conversion': False, # ask
               'find_ignorecase': False,
               'grep_ignorecase': True,
               'grep_regex': True}
OPTIONS_TEXT = 'automatic_file_encoding_conversion: {}\nsort_type: {}'.format("never = -1, ask = 0, always = 1", ', '.join([str(s) for s in SortType]))

DEF_CONFIRMATIONS = {'delete': True,
                     'overwrite': True,
                     'quit': True,
                     'ask_rebuild_vfs': True}

DEF_MISC = {'backup_extension': '.bak',
            'diff_type': 'unified'}
MISC_TEXT = 'diff_type: context, unified, ndiff'

DEF_PROGRAMS = {'shell': 'bash',
                'pager': 'less',
                'editor': 'vi',
                'web': 'firefox',
                'audio': 'vlc',
                'video': 'vlc',
                'graphics': 'eog',
                'pdf': 'evince',
                'ebook': 'FBReader'}
FILES_EXT = {'archive': ['gz', 'bz2', 'xz', 'lz', 'lz4', 'tar', 'tgz', 'tbz2',
                         'txz', 'tlz', 'tlz4', 'Z', 'zip', 'rar', '7z', 'arj',
                         'cab', 'lzh', 'lha', 'zoo', 'arc', 'ark',
                         'rpm', 'deb'],
             'audio': ['au', 'flac', 'mid', 'midi', 'mp2', 'mp3', 'mpg', 'ogg', 'wma', 'xm'],
             'data': ['dta', 'nc', 'dbf', 'mdn', 'db', 'mdb', 'dat',
                      'fox', 'dbx', 'mdx', 'sql', 'mssql', 'msql',
                      'ssql', 'pgsql', 'cdx', 'dbi', 'sqlite'],
             'devel': ['c', 'h', 'cc', 'hh', 'cpp', 'hpp',
                       'py', 'pyw', 'hs', 'lua', 'pl', 'pm', 'inc', 'rb',
                       'asm', 'pas', 'f', 'f90', 'pov', 'm', 'pas',
                       'cgi', 'php', 'phps', 'tcl', 'tk',
                       'js', 'java', 'jav', 'jasm', 'vala', 'glade', 'ui',
                       'diff', 'patch', 'css',
                       'sh', 'bash', 'awk', 'm4', 'el',
                       'st', 'mak', 'sl', 'ada', 'caml',
                       'ml', 'mli', 'mly', 'mll', 'mlp', 'prg'],
             'document': ['txt', 'text', 'rtf',
                          'odt', 'odc', 'odp',
                          'abw', 'gnumeric',
                          'sxw', 'sxc', 'sxp', 'sdw', 'sdc', 'sdp',
                          'djvu', 'dvi', 'bib', 'tex',
                          'doc', 'xls', 'ppt', 'pps', 'docx', 'xlsx', 'pptx',
                          'xml', 'xsd', 'xslt', 'sgml', 'dtd',
                          'mail', 'msg', 'letter', 'ics', 'vcs', 'vcard',
                          'lsm', 'po', 'man', '1', 'info'],
             'ebook': ['azw', 'azw3', 'chm', 'epub', 'fb2', 'imp', 'lit', 'mobi', 'prc'],
             'graphics': ['jpg', 'jpeg', 'gif', 'png', 'tif', 'tiff',
                          'pcx', 'bmp', 'xpm', 'xbm', 'eps', 'pic',
                          'rle', 'ico', 'wmf', 'omf', 'ai', 'cdr',
                          'xcf', 'dwb', 'dwg', 'dxf', 'svg', 'dia'],
             'pdf': ['pdf', 'ps'],
             'temp': ['tmp', '$$$', '~', 'bak'],
             'web': ['html', 'shtml', 'htm'],
             'video': ['acc', 'avi', 'asf', 'flv', 'mkv', 'mov', 'mol', 'mpl', 'med',
                       'mp4', 'mpg', 'mpeg', 'ogv', 'swf', 'ogv', 'wmv']}

POWERCLI_FAVS = ['mv "$f" "{$f.replace(\'\', \'\')}"',
                 'less "$f" $',
                 'find "$p" -name "*" -print0 | xargs --null -0 grep -EHcni "TODO|WARNING|FIXME|BUG"',
                 'find "$p" -name "*" -print0 | xargs --null -0 grep -EHcni "TODO|WARNING|FIXME|BUG" >output.txt &',
                 'cp $s "$o"',
                 '',
                 '',
                 '',
                 '',
                 '']


########################################################################
##### Config
class SectContainer:
    def __init__(self, d):
        for k, v in d.items():
            setattr(self, k, v)

    def prepare_to_save(self, comment=''):
        # convert True->1, False->0 except non booleans values
        NO_BOOL = ('sort_type', 'backup_extension', 'diff_type')
        d = dict((k, (v if k in NO_BOOL else (1 if v else 0))) for k, v in self.__dict__.items())
        if comment:
            d['#'] = comment
        return OrderedDict(sorted(d.items(), key=lambda k: k[0]))


class Config:
    """Configuration class for lfm"""

    def __init__(self):
        # fill with default values
        self.options = SectContainer(DEF_OPTIONS)
        self.confirmations = SectContainer(DEF_CONFIRMATIONS)
        self.misc = SectContainer(DEF_MISC)
        self.programs = DEF_PROGRAMS
        self.files_ext = FILES_EXT
        self.bookmarks = dict([(b, '/') for b in BOOKMARKS_KEYS])
        self.powercli_favs = POWERCLI_FAVS

    def load(self):
        if not exists(CONFIG_FILE):
            log.warning('Configuration file "{}" does not exist, using default values'.format(CONFIG_FILE))
            self.save()
            return
        try:
            with open(CONFIG_FILE) as f:
                hdr = f.readline()[:-1]
                if hdr and hdr != CONFIGFILE_HEADER:
                    log.error('Configuration file "{}" looks corrupted, using default values'.format(CONFIG_FILE))
                    return
        except:
            log.critical('Can\'t read configuration file "{}", quitting'.format(CONFIG_FILE))
            return
        log.debug('Load configuration file "{}"'.format(CONFIG_FILE))
        cp = ConfigParser()
        cp.read(CONFIG_FILE)
        if cp.has_section('Options'):
            for k, v in cp.items('Options'):
                if k in DEF_OPTIONS:
                    try:
                        exec('self.options.{} = {}'.format(k, v if k=='sort_type' else int(v)!=0))
                    except:
                        log.warning('CONFIGURATION FILE: Invalid value "{}" for "{}" in Options section'.format(v, k))
                else:
                    log.warning('CONFIGURATION FILE: Unknown "{}" in section Options, ignoring'.format(k))
        else:
            log.warning('CONFIGURATION FILE: No "Options" section, using defaults')
        if cp.has_section('Confirmations'):
            for k, v in cp.items('Confirmations'):
                if k in DEF_CONFIRMATIONS:
                    try:
                        exec('self.confirmations.{} = {}'.format(k, int(v)!=0))
                    except:
                        log.warning('CONFIGURATION FILE: Invalid value "{}" for "{}" in Confirmations section'.format(v, k))
                else:
                    log.warning('CONFIGURATION FILE: Unknown "{}" in section Confirmations, ignoring'.format(k))
        else:
            log.warning('CONFIGURATION FILE: No "Confirmations" section, using defaults')
        if cp.has_section('Misc'):
            for k, v in cp.items('Misc'):
                if k in DEF_MISC:
                    try:
                        if k == 'diff_type' and v not in ('context', 'unified', 'ndiff'):
                            raise ValueError
                        exec('self.misc.{} = "{}"'.format(k, v))
                    except:
                        log.warning('CONFIGURATION FILE: Invalid value "{}" for "{}" in Misc section'.format(v, k))
                else:
                    log.warning('CONFIGURATION FILE: Unknown "{}" in section Misc, ignoring'.format(k))
        else:
            log.warning('CONFIGURATION FILE: No "Misc" section, using defaults')
        if cp.has_section('Programs'):
            for k, v in cp.items('Programs'):
                if k in DEF_PROGRAMS:
                    # Don't check if program exists
                    self.programs[k] = v.strip()
                else:
                    log.warning('CONFIGURATION FILE: Unknown "{}" in section Programs, ignoring'.format(k))
        else:
            log.warning('CONFIGURATION FILE: No "Programs" section, using defaults')
        if cp.has_section('Files'):
            for k, v in cp.items('Files'):
                if k in FILES_EXT:
                    self.files_ext[k] = list(map(str.lower, [e.strip() for e in v.split(',')]))
                else:
                    log.warning('CONFIGURATION FILE: Unknown "{}" key in section Files, ignoring'.format(k))
        else:
            log.warning('CONFIGURATION FILE: No "Files" section, filling with defaults')
        if cp.has_section('Bookmarks'):
            for k, v in cp.items('Bookmarks'):
                if k in BOOKMARKS_KEYS:
                    # Don't check if exists
                    # if not exists(v.split()) or not isdir(v.split()):
                    #     continue
                    self.bookmarks[k] = v.strip()
                else:
                    log.warning('CONFIGURATION FILE: Unknown "{}" key in section Bookmarks, ignoring'.format(k))
        else:
            log.warning('CONFIGURATION FILE: No "Bookmarks" section, filling with defaults')
        if cp.has_section('PowerCLI Favs'):
            for i, (_, v) in enumerate(cp.items('PowerCLI Favs')):
                self.powercli_favs[i] = v.strip()
        else:
            log.warning('CONFIGURATION FILE: No "PowerCLI Favs" section, filling with defaults')

    def save(self):
        log.debug('Save configuration file "{}"'.format(CONFIG_FILE))
        cfg = ConfigParserWithComments(CONFIGFILE_HEADER)
        cfg['Options'] = self.options.prepare_to_save(OPTIONS_TEXT)
        cfg['Confirmations'] = self.confirmations.prepare_to_save()
        cfg['Misc'] = self.misc.prepare_to_save(MISC_TEXT)
        cfg['Programs'] = OrderedDict(sorted(self.programs.items(), key=lambda k: k[0]))
        d = dict((k, ', '.join(sorted(v))) for k, v in self.files_ext.items())
        cfg['Files'] = OrderedDict(sorted(d.items(), key=lambda k: k[0]))
        cfg['Bookmarks'] = OrderedDict(sorted(self.bookmarks.items(), key=lambda k: k[0]))
        cfg['PowerCLI Favs'] = OrderedDict(dict(((i, c) for i, c in enumerate(self.powercli_favs))))
        try:
            with open(CONFIG_FILE, 'w') as cfgfile:
                cfg.write(cfgfile)
        except:
            log.error('Couldn\'t write configuration file "{}"'.format(CONFIG_FILE))
            raise


########################################################################
##### Color Theme
def load_colortheme():
    log.debug('Parse ColorTheme file')
    if not exists(THEME_FILE):
        log.warning('ColorTheme file does not exist, copying default')
        copy_default_colortheme_file()
    cp = ConfigParser()
    cp.read(THEME_FILE)
    if not cp.has_section('Colors'):
        log.warning('ColorTheme file corrupted, copying default')
        copy_default_colortheme_file()
    # parse file
    colors = dict()
    rels = []
    for it, color_desc in cp.items('Colors'):
        it = it.strip().lower()
        if it not in COLOR_ITEMS:
            log.warning('Color item invalid: {}'.format(it))
            continue
        if color_desc[0] == '=':
            toit = color_desc[1:]
            if toit not in COLOR_ITEMS:
                log.warning('Color item pointed to an invalid item: {}: {}'.format(it, toit))
                continue
            rels.append((it, toit))
            continue
        (fg, bg) = map(str.lower, color_desc.split(maxsplit=1))
        if fg not in VALID_COLORS or bg not in VALID_COLORS:
            log.warning('Color item contain invalid colors: {}: {}'.format(it, color_desc))
            continue
        colors[it] = (fg, bg)
    # fill ='s
    for it, toit in rels:
        if toit not in colors:
            log.warning('Color item pointed to undefined item: {}: {}'.format(it, toit))
            continue
        colors[it] = colors[toit]
    # check if missing items
    for it in set(COLOR_ITEMS) - set(colors.keys()):
        log.warning('Missing Color item {}, added as white on black'.format(it))
        colors[it] = ('white', 'black')
    return colors


def copy_default_colortheme_file():
    try:
        with open(THEME_FILE, 'w') as f:
            default_colortheme = get_lfm_data_file_contents('lfm-default.theme')
            f.write(default_colortheme)
    except:
        log.error('Can\'t copy default ColorTheme file: {}'.format(THEME_FILE))
        raise


########################################################################
##### Keys
def load_keys():
    log.debug('Parse Keys file')
    if not exists(KEYS_FILE):
        log.warning('Keys file does not exist, copying default')
        copy_default_keys_file()
    cp = ConfigParser()
    cp.read(KEYS_FILE)
    if not cp.has_section('Main'):
        log.warning('Keys file corrupted, copying default')
        copy_default_keys_file()
    # parse file
    public_api = get_public_actions()
    actions = dict()
    for it, key_desc in cp.items('Main'):
        it = it.strip().lower()
        if it not in public_api:
            log.warning('KEYS: Action item invalid in section Main: {}'.format(it))
            continue
        try:
            actions[it] = [key_str2bin(k) for k in key_desc.split()]
        except:
            log.warning('KEYS: Action "{}" contains invalid key definition in section Main: "{}"'.format(it, key_desc))
    # check if missing items
    for it in set(public_api) - set(actions.keys()):
        log.warning('KEYS: Missing Action item {}'.format(it))
    # transpose dict:  {action: keys} => {keys: action}
    # and check if same key used for different actions
    # keys = dict([(k, a) for a, ks in actions.items() for k in ks])
    keys = dict()
    for a, ks in actions.items():
        for k in ks:
            if k in keys.keys():
                log.warning('KEYS: Can\'t use key "{}" in action "{}", alredy used in "{}"'.format(key_bin2str(k), a, keys[k]))
                continue
            else:
                keys[k] = a
    return keys


def copy_default_keys_file():
    try:
        with open(KEYS_FILE, 'w') as f:
            default_keys = get_lfm_data_file_contents('lfm-default.keys')
            f.write(default_keys)
    except:
        log.error('Can\'t copy default Keys file: {}'.format(KEYS_FILE))
        raise


def dump_keys_to_file(keys):
    log.info('Dump keys into file')
    with open(DUMP_KEYS_FILE, 'w') as f:
        f.write('#'*30 + ' LFM dump keys ' + '#'*30 + '\n\n')
        f.write('#'*20 + ' Key -> Action ' + '#'*20 + '\n')
        lines = ['{:10} ->   {}'.format(key_bin2str(k), a) for k, a in keys.items()]
        f.write('\n'.join(sorted(lines)))
        f.write('\n\n' + '#'*20 + ' Action -> Key ' + '#'*20 + '\n')
        actions = dict()
        for k, a in keys.items():
            if a in actions:
                actions[a].append(key_bin2str(k))
            else:
                actions[a] = [key_bin2str(k)]
        lines = ['{:24} ->  {}'.format(a, ', '.join(sorted(ks))) for a, ks in actions.items()]
        f.write('\n'.join(sorted(lines)))
        missing_actions = set(get_public_actions()) - set(actions)
        if len(missing_actions) > 0:
            f.write('\n\nActions with no keys:\n')
            for a in sorted(missing_actions):
                f.write('. {}\n'.format(a))
        else:
            f.write('\n\nAll actions have key binding\n')
        f.write('\n' + '#'*70)


########################################################################
##### History
class History(dict):
    def __init__(self):
        self._data = HISTORY_BLANK

    def __getitem__(self, key):
        return self._data[key]

    def __setitem__(self, key, val):
        self._data[key] = val

    def delete(self):
        log.info('Delete history')
        self._data = HISTORY_BLANK
        self.save()

    def load(self):
        log.info('Load history file from "{}"'.format(HISTORY_FILE))
        with open(HISTORY_FILE, 'rb') as f:
            self._data = pickle.load(f)

    def save(self):
        log.info('Save history file to "{}"'.format(HISTORY_FILE))
        with open(HISTORY_FILE, 'wb') as f:
            pickle.dump(self._data, f, -1)

    def append(self, section, new_entry):
        assert section in self._data.keys()
        if new_entry == '':
            return
        if new_entry in self._data[section]:
            self._data[section].remove(new_entry)
        self._data[section].append(new_entry)
        self._data[section] = self._data[section][-HISTORY_MAX:]


########################################################################



# -*- coding: utf-8 -*-


from os.path import dirname, basename, join, isfile, isdir

from utils import delete_bulk
from common import *


######################################################################
def get_compressed_file_engine(filename):
    for p in packagers:       # Note: tbz2 must be before bz2
        for e in p.exts:
            if filename.endswith(e):
                return p(filename)
    else:
        return None


def get_compressed_file_type(filename):
    c = get_compressed_file_engine(filename)
    return c.type if c else None


def check_compressed_vfs(filename):
    c = get_compressed_file_engine(filename)
    return c.can_vfs if c else False


class PackagerBase:
    def __init__(self, filename):
        self.path = filename
        self.dirname = dirname(self.path)
        self.filename = basename(self.path)

    @property
    def cmd_uncompress(self):
        return self.uncompress_cmd % self.path

    @property
    def cmd_compress(self):
        newfile = self.filename + self.exts[0]
        if isfile(self.path):
            if self.type in ('bz2', 'gz', 'xz', 'lz', 'lz4'):
                return self.compress_cmd % self.filename
            elif self.type in ('tbz2', 'tgz', 'txz', 'tlz', 'tlz4, ''tar'):
                return # Don't use tar, it's a file
            else:
                return self.compress_cmd % (self.filename, newfile)
        elif isdir(self.path):
            if self.type in ('bz2', 'gz', 'xz', 'lz', 'lz4'):
                return # Don't compress without tar, it's a dir
            if self.need_tar:
                return self.compress_cmd % (self.filename, newfile)
            else:
                return self.compress_cmd % (newfile, self.filename)

    def cmd_compress2(self, src, dest):
        if not dest.endswith(self.exts[0]):
            dest += self.exts[0]
        if self.need_tar:
            return self.compress2_cmd % (src, dest)
        else:
            return self.compress2_cmd % (dest, src)

    def delete_temp(self, path, from_compress, is_tmp=False):
        if is_tmp:
            tmpfile = path
        else:
            if from_compress:
                for e in self.exts:
                    if self.filename.endswith(e):
                        dirname = self.filename[:-len(e)]
                        break
                else:
                    return
                tmpfile = join(path, dirname)
            else: # from uncompress
                tmpfile = join(path, self.filename+self.exts[0])
        delete_bulk(str(tmpfile), ignore_errors=True)


class PackagerTBZ2(PackagerBase):
    type = 'tbz2'
    exts = ('.tar.bz2', '.tbz2')
    need_tar = True
    can_vfs = True
    uncompress_prog = compress_prog = SYSPROGS['bzip2']
    uncompress_cmd = uncompress_prog + ' -d \"%s\" -c | ' + SYSPROGS['tar'] + ' xfi -'
    compress_cmd = SYSPROGS['tar'] + ' cf - \"%s\" | ' + compress_prog + ' > \"%s\"'
    compress2_cmd = SYSPROGS['tar'] + ' cf - %s | ' + compress_prog + ' > \"%s\"'


class PackagerBZ2(PackagerBase):
    type = 'bz2'
    exts = ('.bz2', )
    need_tar = False
    can_vfs = False
    uncompress_prog = compress_prog = SYSPROGS['bzip2']
    uncompress_cmd = uncompress_prog + ' -d \"%s\"'
    compress_cmd = compress_prog + ' \"%s\"'
    compress2_cmd = compress_prog + ' %s'


class PackagerTGZ(PackagerBase):
    type = 'tgz'
    exts = ('.tar.gz', '.tgz', '.tar.Z')
    need_tar = True
    can_vfs = True
    uncompress_prog = compress_prog = SYSPROGS['gzip']
    uncompress_cmd = uncompress_prog + ' -d \"%s\" -c | ' + SYSPROGS['tar'] + ' xf -'
    compress_cmd = SYSPROGS['tar'] + ' cf - \"%s\" | ' + compress_prog + ' > \"%s\"'
    compress2_cmd = SYSPROGS['tar'] + ' cf - %s | ' + compress_prog + ' > \"%s\"'


class PackagerGZ(PackagerBase):
    type = 'gz'
    exts = ('.gz', )
    need_tar = False
    can_vfs = False
    uncompress_prog = compress_prog = SYSPROGS['gzip']
    uncompress_cmd = uncompress_prog + ' -d \"%s\"'
    compress_cmd = compress_prog + ' \"%s\"'
    compress2_cmd = compress_prog + ' %s'


class PackagerTXZ(PackagerBase):
    type = 'txz'
    exts = ('.tar.xz', '.txz')
    need_tar = True
    can_vfs = True
    uncompress_prog = compress_prog = SYSPROGS['xz']
    uncompress_cmd = uncompress_prog + ' -d \"%s\" -c | ' + SYSPROGS['tar'] + ' xf -'
    compress_cmd = SYSPROGS['tar'] + ' cf - \"%s\" | ' + compress_prog + ' > \"%s\"'
    compress2_cmd = SYSPROGS['tar'] + ' cf - %s | ' + compress_prog + ' > \"%s\"'


class PackagerXZ(PackagerBase):
    type = 'xz'
    exts = ('.xz', )
    need_tar = False
    can_vfs = False
    uncompress_prog = compress_prog = SYSPROGS['xz']
    uncompress_cmd = uncompress_prog + ' -d \"%s\"'
    compress_cmd = compress_prog + ' \"%s\"'
    compress2_cmd = compress_prog + ' %s'


class PackagerTLZ(PackagerBase):
    type = 'tlz'
    exts = ('.tar.lz', '.tlz')
    need_tar = True
    can_vfs = True
    uncompress_prog = compress_prog = SYSPROGS['lzip']
    uncompress_cmd = uncompress_prog + ' -d \"%s\" -c | ' + SYSPROGS['tar'] + ' xf -'
    compress_cmd = SYSPROGS['tar'] + ' cf - \"%s\" | ' + compress_prog + ' > \"%s\"'
    compress2_cmd = SYSPROGS['tar'] + ' cf - %s | ' + compress_prog + ' > \"%s\"'


class PackagerLZ(PackagerBase):
    type = 'lz'
    exts = ('.lz', )
    need_tar = False
    can_vfs = False
    uncompress_prog = compress_prog = SYSPROGS['lzip']
    uncompress_cmd = uncompress_prog + ' -d \"%s\"'
    compress_cmd = compress_prog + ' \"%s\"'
    compress2_cmd = compress_prog + ' %s'


class PackagerTLZ4(PackagerBase):
    type = 'tlz4'
    exts = ('.tar.lz4', '.tlz4')
    need_tar = True
    can_vfs = True
    uncompress_prog = compress_prog = SYSPROGS['lz4']
    uncompress_cmd = uncompress_prog + ' -q -d \"%s\" -c | ' + SYSPROGS['tar'] + ' xf -'
    compress_cmd = SYSPROGS['tar'] + ' cf - \"%s\" | ' + compress_prog + ' -9 -q > \"%s\"'
    compress2_cmd = SYSPROGS['tar'] + ' cf - %s | ' + compress_prog + ' -9 -q > \"%s\"'


class PackagerLZ4(PackagerBase):
    type = 'lz4'
    exts = ('.lz4', )
    need_tar = False
    can_vfs = False
    uncompress_prog = compress_prog = SYSPROGS['lz4']
    uncompress_cmd = uncompress_prog + ' --rm -m -q -d \"%s\"'
    compress_cmd = compress_prog + ' --rm -m -q \"%s\"'
    compress2_cmd = compress_prog + ' --rm -m -q %s'


class PackagerTAR(PackagerBase):
    type = 'tar'
    exts = ('.tar', )
    need_tar = False
    can_vfs = True
    uncompress_prog = compress_prog = SYSPROGS['tar']
    uncompress_cmd = uncompress_prog + ' xf \"%s\"'
    compress_cmd = compress_prog + ' cf \"%s\" \"%s\"'
    compress2_cmd = compress_prog + ' cf \"%s\" %s'


class PackagerZIP(PackagerBase):
    type = 'zip'
    exts = ('.zip', '.jar', '.apk')
    need_tar = False
    can_vfs = True
    uncompress_prog = SYSPROGS['unzip']
    uncompress_cmd = uncompress_prog + ' -o -q \"%s\"'
    compress_prog = SYSPROGS['zip']
    compress_cmd = compress_prog + ' -qr \"%s\" \"%s\"'
    compress2_cmd = compress_prog + ' -qr \"%s\" %s'


class PackagerRAR(PackagerBase):
    type = 'rar'
    exts = ('.rar', )
    need_tar = False
    can_vfs = True
    uncompress_prog = compress_prog = SYSPROGS['rar']
    uncompress_cmd = uncompress_prog + ' x \"%s\"'
    compress_cmd = compress_prog + ' a \"%s\" \"%s\"'
    compress2_cmd = compress_prog + ' a \"%s\" %s'


class Packager7Z(PackagerBase):
    type = '7z'
    exts = ('.7z', )
    need_tar = False
    can_vfs = True
    uncompress_prog = SYSPROGS['7z']
    uncompress_cmd = uncompress_prog + ' x \"%s\"'
    compress_prog = SYSPROGS['7z']
    compress_cmd = compress_prog + ' a \"%s\" \"%s\"'
    compress2_cmd = compress_prog + ' a \"%s\" %s'


######################################################################
packagers = (PackagerTBZ2, PackagerBZ2,
             PackagerTGZ, PackagerGZ,
             PackagerTXZ, PackagerXZ,
             PackagerTLZ, PackagerLZ,
             PackagerTLZ4, PackagerLZ4,
             PackagerTAR, PackagerZIP,
             PackagerRAR, Packager7Z)

packagers_by_type = {'tbz2': PackagerTBZ2,
                     'bz2': PackagerBZ2,
                     'tgz': PackagerTGZ,
                     'gz': PackagerGZ,
                     'txz': PackagerTXZ,
                     'xz': PackagerXZ,
                     'tlz': PackagerTLZ,
                     'lz': PackagerLZ,
                     'tlz4': PackagerTLZ4,
                     'lz4': PackagerLZ4,
                     'tar': PackagerTAR,
                     'zip': PackagerZIP,
                     'rar': PackagerRAR,
                     '7z': Packager7Z}


######################################################################



# -*- coding: utf-8 -*-


import curses
from string import ascii_letters, digits, punctuation
from common import *


########################################################################
VALID_KEYS_1 = ascii_letters + digits + punctuation
VALID_KEYS_2 = "up down left right pageup pagedown home end del ins backspace enter spc esc tab btab".split()
VALID_KEYS_3 = ['f%d' % i for i in range(1, 25)]

tbl_kstr_kcode = [('up', curses.KEY_UP),
                  ('down', curses.KEY_DOWN),
                  ('left', curses.KEY_LEFT),
                  ('right', curses.KEY_RIGHT),
                  ('pageup', curses.KEY_PPAGE),
                  ('pagedown', curses.KEY_NPAGE),
                  ('home', curses.KEY_HOME),
                  ('end', curses.KEY_END),
                  ('del', curses.KEY_DC),
                  ('ins', curses.KEY_IC),
                  ('backspace', curses.KEY_BACKSPACE),
                  ('enter', 10),
                  ('spc', 32),
                  ('esc', 27),
                  ('tab', 9),
                  ('backtab', curses.KEY_BTAB),
                  ('S+up', curses.KEY_SR),
                  ('S+down', curses.KEY_SF),
                  ('S+left', curses.KEY_SLEFT),
                  ('S+right', curses.KEY_SRIGHT),
                  ('S+del', curses.KEY_SDC)]
kmap_str2code = dict(tbl_kstr_kcode)
kmap_code2str = dict([(it[1], it[0]) for it in tbl_kstr_kcode])

# A-up: kUP3, A-S-up: kUP4, C-up: kUP5, C-S-up: kUP6, C-A-up: kUP7 ... C-A-pagedown: kNXT7
# kUP3: 0x235, kUP4: 0x236 ... kDN3: 0x20c ... kLFT3: 0x220 ... kRIT3: 0x22c ...
# kDC3: 0x206 ... kIC3: 0x21b ... kPRV: 0x22a ... kNXT: 0x22a
tbl_mods = {'A-': 3, 'A-S-': 4, 'S-A-': 4, 'C-': 5, 'C-S-': 6, 'S-C-': 6, 'C-A-': 7, 'A-C-': 7}
#tbl_keysspecial = {'up': ('kUP', 0x235), 'down': ('kDN', 0x20c), 'left': ('kLFT', 0x220),
#                   'right': ('kRIT', 0x22f), 'del': ('kDC', 0x206), 'ins': ('kIC', 0x21b),
#                   'pageup': ('kPRV', 0x22a), 'pagedown': ('kNXT', 0x225)}
tbl_keysspecial = {'up': ('kUP', 0x236), 'down': ('kDN', 0x20d), 'left': ('kLFT', 0x221),
                   'right': ('kRIT', 0x230), 'del': ('kDC', 0x207), 'ins': ('kIC', 0x21c),
                   'pageup': ('kPRV', 0x22b), 'pagedown': ('kNXT', 0x226)}
# tbl_keysspecial = {'up': ('kUP', 0x237), 'down': ('kDN', 0x20e), 'left': ('kLFT', 0x222),
#                    'right': ('kRIT', 0x231), 'del': ('kDC', 0x208), 'ins': ('kIC', 0x21d),
#                    'pageup': ('kPRV', 0x22c), 'pagedown': ('kNXT', 0x227)}
tbl_aliases = []
for mod in tbl_mods:
    for k in tbl_keysspecial:
        pair = mod+k, tbl_keysspecial[k][1]+tbl_mods[mod]-3
        tbl_aliases.append(pair)
# S-F1: F13 ... S-F12: F24
for i in range(0, 11):
    pair = 'S-F{}'.format(i+1), curses.KEY_F1+12+i
    tbl_aliases.append(pair)
kmap_aliases_str2code = dict(tbl_aliases)
kmap_aliases_code2str = dict([(it[1], it[0]) for it in tbl_aliases])


########################################################################
def get_keycode(kstr):
    if len(kstr) == 1:
        if kstr in VALID_KEYS_1:
            return ord(kstr)
        else:
            raise ValueError
    kstr = kstr.lower()
    if kstr in kmap_aliases_str2code:
        return kmap_aliases_str2code[kstr]
    if kstr in VALID_KEYS_2:
        return kmap_str2code[kstr]
    if kstr in VALID_KEYS_3:
        return curses.KEY_F1 + int(kstr[1:]) - 1
    else:
        raise ValueError

def key_str2bin(k):
    if k in kmap_aliases_str2code:
        return (KeyModifier.none, kmap_aliases_str2code[k])
    if k.startswith('C-'):
        km = KeyModifier.control
        k = k[2:]
        if len(k) == 1:
            return (KeyModifier.none, ord(k.lower())-ord('a')+1)
    elif k.startswith('A-'):
        km = KeyModifier.alt
        k = k[2:]
    else:
        km = KeyModifier.none
    return (km, get_keycode(k))


def get_keystr(kcode):
    if kcode in kmap_code2str:
        return kmap_code2str[kcode]
    if kcode in kmap_aliases_code2str:
        return kmap_aliases_code2str[kcode]
    if kcode >= curses.KEY_F1 and kcode <= curses.KEY_F24:
        return 'f%d' % (kcode-curses.KEY_F1+1, )
    if kcode>=32 and kcode<128:
        return chr(kcode)
    return str(kcode)

def key_bin2str(k):
    km, key_code = k
    if key_code < 0x20:
        if key_code in kmap_code2str:
            return kmap_code2str[key_code]
        return 'C-' + chr(ord('a')-1+key_code)
    if km == KeyModifier.alt:
        km = 'A-'
    else:
        km = ''
    return km + get_keystr(key_code)


########################################################################



# -*- coding: utf-8 -*-


import logging
import os.path
from enum import Enum, IntEnum
from string import digits, ascii_lowercase


########################################################################
##### General
AUTHOR = 'Iñigo Serna'
VERSION = '3.0'
DATE = '2001-17'
LFM_NAME = 'lfm - Last File Manager'

CONFIG_DIR = os.path.abspath(os.path.expanduser('~/.config/lfm'))
CONFIG_FILE = os.path.join(CONFIG_DIR, 'lfm.ini')
THEME_FILE = os.path.join(CONFIG_DIR, 'lfm.theme')
KEYS_FILE = os.path.join(CONFIG_DIR, 'lfm.keys')
HISTORY_FILE = os.path.join(CONFIG_DIR, 'lfm.history')
DUMP_KEYS_FILE = os.path.join(CONFIG_DIR, 'keys.dump')

DEBUG_LEVEL = logging.INFO # DEBUG, WARNING
log = logging.getLogger('lfm')

MAX_TABS = 4
VFS_STRING = '#vfs://'
BOOKMARKS_KEYS = digits + ascii_lowercase
MIN_COLUMNS = 66

HISTORY_MAX = 100
HISTORY_BLANK = {'file': [], 'path': [], 'glob': [], 'find': [], 'grep': [], 'exec': [], 'cli': []}

SYSPROGS = {'tar': 'tar',
            'bzip2': 'bzip2',
            'gzip': 'gzip',
            'zip': 'zip',
            'unzip': 'unzip',
            'rar': 'unrar',
            '7z': '7z',
            'xz': 'xz',
            'lzip': 'lzip',
            'lz4': 'lz4',
            'grep': 'grep',
            'find': 'find',
            'which': 'which',
            'xargs': 'xargs'}


FileType = IntEnum('FileType', 'dir link2dir link nlink cdev bdev fifo socket exe reg unknown')
FILETYPES = [('x', 'Placeholder'),
             ('/', 'Directory'),
             ('~', 'Link to Directory'),
             ('@', 'Link'),
             ('!', 'No Link'),
             ('-', 'Char Device'),
             ('+', 'Block Device'),
             ('|', 'Fifo'),
             ('#', 'Socket'),
             ('*', 'Executable'),
             (' ', 'File'),
             ('?', 'Unknown')]

PaneMode = Enum('PaneMode', 'full half hidden info contents')
SortType = Enum('SortType', 'none byName byExt byPath bySize byMTime')
KeyModifier = Enum('KeyModifier', 'none control alt')
RetCode = Enum('RetCode', 'nothing quit_chdir quit_nochdir fix_limits full_redisplay half_redisplay')

ProcCode = Enum('ProcCode', 'end stopped error next confirm')
ProcCodeConfirm = Enum('ProcCodeConfirm', 'stop ok skip')

class LFMFileExistsError(Exception):
    pass

class LFMFileSkipped(Exception):
    pass

class LFMTerminalTooNarrow(Exception):
    pass


########################################################################
##### Color Theme
COLOR_ITEMS = [
    # general interface
    'header', 'tab_active', 'tab_inactive', 'pane_active', 'pane_inactive',
    'pane_header_path', 'pane_header_titles',
    'statusbar', 'powercli_prompt', 'powercli_text', 'selected_files', 'cursor', 'cursor_selected',
    # files. Must match the FILES_EXT from configuration!
    'files_dir', 'files_exe', 'files_reg', 'files_archive', 'files_audio', 'files_data',
    'files_devel', 'files_document', 'files_ebook', 'files_graphics', 'files_pdf',
    'files_temp', 'files_video', 'files_web',
    # dialogs
    'dialog', 'dialog_title', 'button_active', 'button_inactive',
    'dialog_error', 'dialog_error_title', 'dialog_error_text', 'dialog_perms',
    'selectitem', 'selectitem_title', 'selectitem_cursor', 'entryline',
    'progressbar_fg', 'progressbar_bg',
    'view_white_on_black', 'view_red_on_black', 'view_blue_on_black', 'view_green_on_black'
]

VALID_COLORS = ['white', 'black', 'red', 'green', 'yellow', 'blue', 'magenta', 'cyan',
                'white*', 'black*', 'red*', 'green*', 'yellow*', 'blue*', 'magenta*', 'cyan*']


########################################################################
